


# Overview

**Paper** : Optimal Prediction of Synchronization-Reversal Free Races, POPL'21

**Artifact Outline** - In the above paper, we define a new notion of synchronization-reversal-free data races, and present an algorithm *SRFree* for detecting them.
We have implemented SRFree in the tool RAPID, and compared its performance against two other algorithms in RAPID, namely *Schedulably Happens Before (SHB)* and *Weakly Causally Precedes (WCP)*, as well as against the algorithm *M2*.

**REMARK:** Following review suggestions, our final submission will feature a small title change, to
```
Optimal Prediction of Synchronization-Preserving Races
```
For this reason, all references to our algorithm below and in our tool have been changed from SRFree to SyncPreserving




# Claims
The current artifact allows to reproduce the experimental results of the paper, namely Tables 1, 2, and 3.
We remark that the tables were produced by running the corresponding benchmarks on a cluster.
If running on a cluster is not possible, we give some examples below on how to produce selective parts of these tables, by running the more light benchmarks.


# Download, installation, and sanity-testing

### 1. Requirements

1. Machine with Unix based operating system, such as Ubuntu or MacOS. Our experiments were performed on a  machine with about 30GB RAM. Our experiments work even with less RAM but the performance might differ.
2. Java 1.8 or higher (required for executing RAPID and M2)
3. apache ant (required for building RAPID)
4. Python 2.7 or higher (required for running the scripts that drive benchmark execution)
5. pdflatex (required for producing the tables)


### 2. Artifact
The artifact is the directory that contains this file, and the structure is described below.
```
AE/
|--- README.md			This README 
|--- m2/				The binaries of the tool M2
|--- paper/				A copy of our paper
|--- rapid-master/		The source and binaries of the tool RAPID, containing our SyncPreserving algorithm
|--- results/			A directory where the results of the executions are stored
|--- scripts/			Various scripts for generating the results in the paper
|--- traces/			The directory with the input traces (to be downloaded)
```

The artifact can also be downloaded from the permanent link
```
https://doi.org/10.5281/zenodo.4073388
```

### 3. Getting/generating the traces
The traces that serve as inputs to the compared methods are large files and are not supplied directly. The traces folder contains a single toy trace *fig1_1.std*, which is the trace of Figure 1(a) in the paper, used for test runs.

 For the traces used as benchmarks in Tables 1,2 and 3, download the .std files from
```
https://drive.google.com/drive/folders/1qlSiSIJihq8wqldh9f4a_5vjl1AowZlc
```
The url contains a directory *compressed_logs* and a file *syn-preserving-logs.tar.gz*
The former contains one compressed trace per benchmark, while the latter contains one compressed directory with all traces.
In each case, the compressed files must be extracted to obtain the corresponding *.std* files before proceeding.

**REMARK:** Extracting *sync-preserving-logs.tar.gz* on a 2019 MacBook Pro takes about 5 minutes, and after extraction the folder is about 61GBs in size.
When uncompressed, the largest trace is luindex.std, about 23GBs.
If there is lack of space, one can download selectively some benchmarks from the *compressed_logs* directory and perform experiments on those (as we explain below).
We expect that all benchmarks up to *jigsaw* from Table 1 can be run on a laptop, with the exception of M2 on linkedlist, which is likely to time out.


 Each trace file is a text file ready to be processed by RAPID. 
 M2 takes a different format. To generate a trace readable by M2, execute
```
python scripts/std2m2.py <tracefile or tracedir>
```
e.g.,
```
python scripts/std2m2.py tracefile.std 			(for conversion of one file)
python scripts/std2m2.py traces/ 				(for conversion of all files in the traces directory)
```
The commands above convert a tracefile.std (read by RAPID) to tracefile.m2 (read by M2).
Both file types are human-readable.
If (in later steps) one performs runs using M2, the corresponding benchmark must have been converted to the *.m2* format using the command above.

**REMARK:** Conversion of large traces can take several minutes. E.g., converting xalan.std in a 2016 MacBook Pro with 16GB of memory takes about 15 minutes.

### 4. Building Rapid
The tool M2 comes compiled in a jar file. 
RAPID has to be built, following the two commands below.
```
cd rapid-master
ant jar
```
### 5. Sanity testing

For a sanity test, execute the following, which runs our SyncPreserving algorithm on the input trace of Figure 1(a) in the paper.
```
java  -cp "rapid-master/rapid.jar:rapid-master/lib/*" SyncPreserving -f=std  -p traces/fig1_1.std
```
The output should look similar to the following:
```
Time for Phase-1 = 2 milliseconds
5
Analysis complete
Number of 'racy' events found = 1
Number of 'racy' variables = 1
Max race distance = 5
Avg. race distance = 5.0
Time for analysis = 18 milliseconds
```




# Evaluation instructions
Finally, we describe how to reproduce the claims in Tables 1,2 and 3 of the paper. This is a two step process.

1. Run the compared algorithms (SyncPreserving, WCP, SHB, M2) on the selected traces. Every run of an algorithm A on a trace B produces a file results/result_A_B.txt.
2. Execute a script that reads the results in the results/ directory and produces the corresponding table.

As some traces are very large, they might not be runnable on a laptop.
In that case, one can run selectively some algorithms on some traces, and step 2 above will produce the corrosponding table for only the benchmarks which have a result file in the results/ directory.

We expect that all benchmarks up to *jigsaw* from Table 1 can be run on a laptop, with the exception of M2 on linkedlist, which is likely to time out.


### 1. Running the algorithms
The script run.py takes as a parameter an algorithm and a benchmark and executes the algorithm on the trace.
There is no output printed on the screen.
Whenever an algorithm A is executed on a trace B, a file results/result_A_B.txt is generated.

```
python scripts/run.py <SyncPreserving, SHB, WCP, M2> <tracename>
```
e.g.,
```
python scripts/run.py SyncPreserving wronglock
python scripts/run.py WCP clean
python scripts/run.py M2 moldyn
```
One can also give a comma-separated list of algorithms, and a comma-separated list of trace names all of which will be run on the same benchmark, e.g.,
```
python scripts/run.py SyncPreserving,SHB,WCP,M2 wronglock,clean,moldyn
```
Finally, the second parameter can be a directory of trace files, all of which will be run on the specified algorithms. So to run all algorithms on all benchmarks, one can give
```
python scripts/run.py SyncPreserving,SHB,WCP,M2 traces/
```

**REMARK:** Each execution of an algorithm on some trace times-out after 1h. This parameter can be set in variable *timeout* inside scripts/run.py

**REMARK:** Some traces require large memory and running time. Tables 1,2 and 3 were generated by running the experiments on a cluster with at least 30GB of memory, and it took about 17 hours (taken mostly by 1-hour timeouts of M2 -- Table 1 shows what the expected time is per algorithm and benchmark).
If access on a cluster is not possible, we expect that running all benchmarks up to *jigsaw* of Table 1 should be possible on a modern laptop, and should take no more than 30 minutes + 1 hour for M2 to timeout on benchmark *linkedlist*.



### 2. Generating the tables
Finally, we describe how to generate the 3 tables in the paper.
There is one script scripts/tablei.py for each case (i=1,2,3).
The script will produce the statistics found in the respective table, for as many benchmarks as the result files appearing in the results/ directory.
Missing entries will be marked with *-*.

Table 1 reports race results on all benchmarks, and table1.py takes no parameters.


```
python scripts/table1.py
```

**REMARK:** The last line of Table 1 in the paper contains the totals of running times and race reports for each method. For cases where a timeout had occurred, 1h was added to the total running time of the respective algorithm.
To allow for partial re-production of the tables, here the timeouts are not added towards the total running time.

Tables 2 and 3 report results on racy variables and race distances, respectively, on a selected set of benchmarks.
Running the respective script without parameters will produce the corresponding table, filtering for the bechmarks appear in the paper.

```
python scripts/table2.py
python scripts/table3.py
```

Giving the parameter *all* will produce the statistics of Tables 2 and 3 for all benchmarks present in the results/ folder (i.e., without filtering out the ones not appearing in the paper).

```
python scripts/table2.py all
python scripts/table3.py all
```