from __future__ import division
import os, sys, glob, shlex
from subprocess import Popen, PIPE
from threading import Timer

# Timeout in seconds
timeout = 3600

# JVM memory
jvm_memory = '30000m'

# traces directory
trace_dir = 'traces/'

# results directory
results_dir = 'results/'


# Execute cmd, write output to outfile. Times out after timeout seconds, if so it removes outfile
def timed_run(cmd):
  print cmd
  proc = Popen(cmd, shell=True, stdout=PIPE, stderr=PIPE)
  timer = Timer(timeout, proc.kill)
  try:
    timer.start()
    stdout, stderr = proc.communicate()
  except:
    print 'Timed out after ' + str(timeout) + 'seconds'
  finally:
    timer.cancel()
  
  return stdout



# Execute M2 in tracename
def run_m2(tracename):

  outfile = results_dir + 'result_M2_' + tracename  + '.txt'

  cmd = 'java ' + ' -Xmx' + jvm_memory + ' -Xms' + jvm_memory + ' -cp "m2/m2.jar:m2/lib/*" '
  cmd += ' Main analyze ' + trace_dir + tracename + '.m2' + ' true false ' + outfile
  timed_run(cmd)


# Execute Rapid in tracefile using algo
def run_rapid(tracename, algo):

  outfile = results_dir + 'result_' + algo + '_' + tracename  + '.txt'

  cmd = 'java ' + ' -cp "rapid-master/rapid.jar:rapid-master/lib/*" ' + algo + ' -f=std ' + ' -p ' + trace_dir + tracename + '.std '
  
  outstr = timed_run(cmd)
  if 'complete' in outstr[-200:]:
    f = open(outfile, 'w')
    print>>f, outstr
    f.close()



def main():

  algos = sys.argv[1].split(',')
  tracefiles = []

  if os.path.isdir(sys.argv[2]):
    tracefiles_std = [x.split('/')[-1].split('.')[0] for x in glob.glob(sys.argv[2] + '*.std')]
    tracefiles_m2 = [x.split('/')[-1].split('.')[0] for x in glob.glob(sys.argv[2] + '*.m2')]
  else:
    tracefiles_std = sys.argv[2].split(',')
    tracefiles_m2 = tracefiles_std[:]
  
  for algo in algos:
    if algo == 'M2':
      for tracefile in tracefiles_m2: run_m2(tracefile)
    else:
      for tracefile in tracefiles_std: run_rapid(tracefile, algo)


main()