from util import *

table_header = '''
\\setcounter{table}{0}
\\begin{table}
\\caption{
Dynamic race reports in our benchmarks.
$\\NumEvents$ and $\\NumThreads$ denote the number of events and number of threads in the respective trace.
For races, an entry $r$ ($s$) denotes the number $r$ of events $e_2$ found to be in race with an earlier event $e_1$,
as well as the number $s$ of unique source-code lines corresponding to such events $e_2$.
Bold-face entries highlight cases where there are sync-reversal free races that are not happens-before races.
}
\\label{tab:races}
\\setlength\\tabcolsep{1.5pt}
\\renewcommand{\\arraystretch}{1.2}
\\footnotesize
\\begin{tabular}{|c|c|c|||c|c||c|c||c|c||c|c|}
\hline
\\textbf{Benchmark} & $\\NumEvents$ & $\\NumThreads$ & \\multicolumn{2}{c||}{\\textsf{SHB}}& \\multicolumn{2}{c||}{\\textsf{WCP}}& \\multicolumn{2}{c||}{\\textsf{M2}}& \\multicolumn{2}{c|}{$\\ZeroRevAlgo$}\\\\
\\hline
&&&\\textbf{Races} & \\textbf{Time} &\\textbf{Races} & \\textbf{Time}&\\textbf{Races} & \\textbf{Time}&\\textbf{Races} & \\textbf{Time}\\\\
\\hline
'''

table_footer = '''
\\hline
\\end{tabular}
\\end{table}
'''

def print_table():
	
	totals, total_sizes = {}, 0
	for algo in algos: totals[algo]={'races':0, 'racy_events': 0, 'time': 0}

	results = parse_dir()
	body = ''
	for b in results:
		total_sizes += benchmark_data[b]['size']
		body += '\\texttt{' + short_name(b) + '}' + ' & ' + events_pretty_print(benchmark_data[b]['size']) + ' & ' + str(benchmark_data[b]['threads'])
		for algo in algos:
			body += ' & '
			if 'num_races' in results[b][algo]:
				totals[algo]['races'] += results[b][algo]['num_races']
				totals[algo]['racy_events'] += len(results[b][algo]['racy_events'])
				totals[algo]['time'] += results[b][algo]['time']
				body += str(results[b][algo]['num_races']) + ' (' + str(len(results[b][algo]['racy_events'])) + ')' + ' & ' + time_pretty_print(results[b][algo]['time'])
			else:
				body += ' - & - '
		body += '\\\\\n'
	
	body += '\\hline\n\\textbf{Totals} & ' + events_pretty_print(total_sizes) + ' & - '
	for algo in algos:
		body += ' & ' + str(totals[algo]['races']) + ' (' + str(totals[algo]['racy_events']) + ')' + ' & ' + time_pretty_print(totals[algo]['time'])
	body += '\\\\'

	s = table_header + body + table_footer
	stream_latex(s, 'table1')

print_table()
