from __future__ import division
import os, sys, glob, shlex
from subprocess import Popen, PIPE

# results directory
results_dir = 'results/'

algos = ['SHB', 'WCP', 'M2', 'SyncPreserving']


benchmark_data = {
	'account': {'size':134 , 'threads':5 },
	'airlinetickets': {'size':140 , 'threads':5 },
	'array': {'size':51 , 'threads':4 },
	'boundedbuffer': {'size':332 , 'threads':3 },
	'bubblesort': {'size':4641 , 'threads':13 },
	'bufwriter': {'size':22250 , 'threads':7 },
	'critical': {'size':59 , 'threads':5 },
	'derby': {'size':1378161 , 'threads':5 },
	'ftpserver': {'size':49560 , 'threads':12 },
	'jigsaw': {'size':3086982 , 'threads':12 },
	'lusearch': {'size':217485935 , 'threads':8 },
	'mergesort': {'size':3032 , 'threads':6 },
	'moldyn': {'size':200295 , 'threads':4 },
	'pingpong': {'size':151 , 'threads':7 },
	'raytracer': {'size':15771 , 'threads':4 },
	'xalan': {'size':122539064 , 'threads':7 },
	'batik': {'size':157914164 , 'threads':7 },
	'cryptorsa': {'size':58458889 , 'threads':9 },
	'lufact': {'size':134079047 , 'threads':5 },
	'luindex': {'size':397752134 , 'threads':3 },
	'sor': {'size':606921392 , 'threads':5 },
	'sunflow': {'size':11657057 , 'threads':17 },
	'tsp': {'size':307307992, 'threads':10 },
	'clean': {'size':1299 , 'threads':10 },
	'lang': {'size':6257, 'threads':8 },
	'linkedlist': {'size':1002287 , 'threads':13 },
	'producerconsumer': {'size':658 , 'threads':9 },
	'readerswriters': {'size':11267 , 'threads':6 },
	'twostage': {'size':193 , 'threads':13 },
	'wronglock': {'size':246 , 'threads':23 },
}


def short_name(b):
	d = {
		'airlinetickets': 'airtickets',
		'readerswriters': 'readswrites',
		'boundedbuffer': 'bbuffer',
		'producerconsumer': 'prodcons'
	}
	if b in d:
		return d[b]
	else:
		return b

def RepresentsInt(s):
	try:
		int(s)
		return True
	except ValueError:
		return False


def RepresentsFloat(s):
	try:
		float(s)
		return True
	except ValueError:
		return False
	
def time_pretty_print(s):
	
	m = int(s/60)
	h = int(m/60)
	
	p = ''
	if h>0: p = p + str(h) + 'h'
	if m>0:
		p = p + str(int(m%60)) + 'm'
		p = p + str(int(s%60)) + 's'
	else:
		p = p + str("%.2f" % s) + 's'

	return p
		

def events_pretty_print(s):

	
	if s < 10**3:
		return str(s)
	
	if s < 10**4:
		return ("%.1f" % (s/(10**3))) + 'K'
	
	if s < 10**6:
		return str(int(s/10**3)) + 'K'
	
	if s < 10**7:
		return ("%.1f" % (s/(10**6))) + 'M'
	
	if s < 10**9:
		return str(int(s/10**6)) + 'M'
	
	if s < 10**10:
		return ("%.1f" % (s/(10**9))) + 'B'



def parse_result_m2_one(filename):
  
  f = open(filename)
  lines = f.readlines()
  f.close()

  i = 1
  num_races, racy_events = 0, set()
  while 'Analysis complete' not in lines[i]:

    num_races += 1
    racy_events.add(lines[i].strip())
    i += 1

  time = int(lines[i+2].split()[4])/1000

  return {'num_races': num_races, 'racy_events': racy_events, 'time': time}



def parse_result_rapid_one(filename):

	f = open(filename)
	lines = f.readlines()
	f.close()

	i = 1
	num_races, racy_events = 0, set()
	while 'Analysis complete' not in lines[i]:

		num_races += 1
		racy_events.add(lines[i].strip())
		i +=1

	racy_variables = int(lines[i+2].split('=')[1])
	max_distance = int(lines[i+3].split('=')[1])
	time = int(lines[i+5].split('=')[1].split()[0])/1000
  
    

	return {'num_races': num_races, 'racy_events': racy_events, 'racy_variables': racy_variables, 'max_distance': max_distance, 'time': time}


def parse_dir():

  results = {}
  
  for filename in glob.glob(results_dir + 'result_*.txt'):
    parts = filename.split('_')
    algo, benchmark = parts[-2], parts[-1].split('.')[0]
    if benchmark not in results: results[benchmark] = {'SHB':{}, 'WCP':{}, 'SyncPreserving':{}, 'M2':{}}
    
    if algo == 'M2':
      results[benchmark][algo] = parse_result_m2_one(filename)
    else:
      results[benchmark][algo] = parse_result_rapid_one(filename)
  
  return results



latex_header = '''
\\documentclass{article}
\\usepackage[utf8]{inputenc}
\\usepackage[margin=2in]{geometry}
\\begin{document}
\\newcommand{\\NumThreads}{\\mathcal{T}}
\\newcommand{\\NumEvents}{\\mathcal{N}}
\\newcommand{\\ZeroRevAlgo}{\\textsf{SRFree}}
'''

latex_footer = '''
\\end{document}
'''

def stream_latex(s, latex_filename):

	latex_text = latex_header + s + latex_footer

	f = open(latex_filename + '.tex', 'w')
	print>>f, latex_text
	f.close()

	#print latex_text

	cmd = 'cat ' + latex_filename + '.tex' + ' | pdflatex --jobname=' + latex_filename + ' &>/dev/null' + '; rm ' + latex_filename + '.log ' + latex_filename + '.aux ' + latex_filename + '.tex '
	proc = Popen(cmd, shell=True, stdout=PIPE, stderr=PIPE)

  